@@ -13,20 +13,25 @@
 #include <R_ext/Lapack.h>
 #include <R_ext/BLAS.h>
 
-
 extern "C" {
-  SEXP PGOcc(SEXP y_r, SEXP X_r, SEXP Xp_r, SEXP pocc_r, SEXP pdet_r, 
-             SEXP J_r, SEXP nObs_r, SEXP K_r, SEXP betaStarting_r, 
-	     SEXP alphaStarting_r, SEXP zStarting_r, SEXP zLongIndx_r, 
+  SEXP PGOcc(SEXP y_r, SEXP X_r, SEXP Xp_r, SEXP XRE_r, SEXP XpRE_r, 
+             SEXP consts_r, SEXP K_r, SEXP nOccRELong_r, SEXP nDetRELong_r, 
+             SEXP betaStarting_r, SEXP alphaStarting_r, 
+             SEXP sigmaSqPsiStarting_r, SEXP sigmaSqPStarting_r, 
+             SEXP betaStarStarting_r, SEXP alphaStarStarting_r, 
+             SEXP zStarting_r, SEXP zLongIndx_r, 
+             SEXP betaStarIndx_r, SEXP betaLevelIndx_r, 
+	     SEXP alphaStarIndx_r, SEXP alphaLevelIndx_r, 
 	     SEXP muBeta_r, SEXP muAlpha_r, SEXP SigmaBeta_r, SEXP SigmaAlpha_r, 
-	     SEXP nSamples_r, SEXP nThreads_r, SEXP verbose_r, SEXP nReport_r, 
-	     SEXP nBurn_r, SEXP nThin_r, SEXP nPost_r, SEXP currChain_r, 
-	     SEXP nChain_r){
+             SEXP sigmaSqPsiA_r, SEXP sigmaSqPsiB_r, SEXP sigmaSqPA_r, 
+             SEXP sigmaSqPB_r, SEXP nSamples_r, SEXP nThreads_r, 
+             SEXP verbose_r, SEXP nReport_r, SEXP samplesInfo_r,
+	     SEXP chainInfo_r){
    
     /**********************************************************************
      * Initial constants
      * *******************************************************************/
-    int i, j, s, info, nProtect=0;
+    int i, j, s, r, l, info, nProtect=0;
     const int inc = 1;
     const double one = 1.0;
     const double zero = 0.0;
@@ -40,8 +45,17 @@ extern "C" {
     double *y = REAL(y_r);
     double *X = REAL(X_r);
     double *Xp = REAL(Xp_r);
-    int pOcc = INTEGER(pocc_r)[0];
-    int pDet = INTEGER(pdet_r)[0];
+    int *XRE = INTEGER(XRE_r); 
+    int *XpRE = INTEGER(XpRE_r); 
+    // Load constants
+    int J = INTEGER(consts_r)[0];
+    int nObs = INTEGER(consts_r)[1]; 
+    int pOcc = INTEGER(consts_r)[2];
+    int pOccRE = INTEGER(consts_r)[3];
+    int nOccRE = INTEGER(consts_r)[4];
+    int pDet = INTEGER(consts_r)[5];
+    int pDetRE = INTEGER(consts_r)[6];
+    int nDetRE = INTEGER(consts_r)[7];
     int ppDet = pDet * pDet;
     int ppOcc = pOcc * pOcc; 
     double *muBeta = (double *) R_alloc(pOcc, sizeof(double));   
@@ -52,19 +66,27 @@ extern "C" {
     F77_NAME(dcopy)(&ppOcc, REAL(SigmaBeta_r), &inc, SigmaBetaInv, &inc);
     double *SigmaAlphaInv = (double *) R_alloc(ppDet, sizeof(double));   
     F77_NAME(dcopy)(&ppDet, REAL(SigmaAlpha_r), &inc, SigmaAlphaInv, &inc);
-    int J = INTEGER(J_r)[0];
+    double *sigmaSqPsiA = REAL(sigmaSqPsiA_r); 
+    double *sigmaSqPsiB = REAL(sigmaSqPsiB_r); 
+    double *sigmaSqPA = REAL(sigmaSqPA_r); 
+    double *sigmaSqPB = REAL(sigmaSqPB_r); 
+    int *nOccRELong = INTEGER(nOccRELong_r); 
+    int *nDetRELong = INTEGER(nDetRELong_r); 
     double *K = REAL(K_r); 
     int *zLongIndx = INTEGER(zLongIndx_r); 
-    int nObs = INTEGER(nObs_r)[0];
+    int *alphaStarIndx = INTEGER(alphaStarIndx_r); 
+    int *alphaLevelIndx = INTEGER(alphaLevelIndx_r);
+    int *betaStarIndx = INTEGER(betaStarIndx_r); 
+    int *betaLevelIndx = INTEGER(betaLevelIndx_r);
     int nSamples = INTEGER(nSamples_r)[0];
+    int nBurn = INTEGER(samplesInfo_r)[0]; 
+    int nThin = INTEGER(samplesInfo_r)[1];
+    int nPost = INTEGER(samplesInfo_r)[2]; 
+    int currChain = INTEGER(chainInfo_r)[0];
+    int nChain = INTEGER(chainInfo_r)[1];
     int nThreads = INTEGER(nThreads_r)[0];
     int verbose = INTEGER(verbose_r)[0];
     int nReport = INTEGER(nReport_r)[0];
-    int nThin = INTEGER(nThin_r)[0]; 
-    int nBurn = INTEGER(nBurn_r)[0]; 
-    int nPost = INTEGER(nPost_r)[0]; 
-    int currChain = INTEGER(currChain_r)[0];
-    int nChain = INTEGER(nChain_r)[0];
     int status = 0; 
     int thinIndx = 0;
     int sPost = 0;  
@@ -90,7 +112,7 @@ extern "C" {
         Rprintf("Samples per Chain: %i \n", nSamples);
         Rprintf("Burn-in: %i \n", nBurn); 
         Rprintf("Thinning Rate: %i \n", nThin); 
-	Rprintf("Number of Chains: %i \n", nChain);
+        Rprintf("Number of Chains: %i \n", nChain);
         Rprintf("Total Posterior Samples: %i \n\n", nPost * nChain); 
 #ifdef _OPENMP
         Rprintf("\nSource compiled with OpenMP support and model fit using %i thread(s).\n\n", nThreads);
@@ -110,9 +132,21 @@ extern "C" {
     // Occupancy covariates
     double *beta = (double *) R_alloc(pOcc, sizeof(double));   
     F77_NAME(dcopy)(&pOcc, REAL(betaStarting_r), &inc, beta, &inc);
+    // Occupancy random effect variances
+    double *sigmaSqPsi = (double *) R_alloc(pOccRE, sizeof(double)); 
+    F77_NAME(dcopy)(&pOccRE, REAL(sigmaSqPsiStarting_r), &inc, sigmaSqPsi, &inc); 
+    // Latent occupancy random effects
+    double *betaStar = (double *) R_alloc(nOccRE, sizeof(double)); 
+    F77_NAME(dcopy)(&nOccRE, REAL(betaStarStarting_r), &inc, betaStar, &inc); 
     // Detection covariates
     double *alpha = (double *) R_alloc(pDet, sizeof(double));   
     F77_NAME(dcopy)(&pDet, REAL(alphaStarting_r), &inc, alpha, &inc);
+    // Detection random effect variances
+    double *sigmaSqP = (double *) R_alloc(pDetRE, sizeof(double)); 
+    F77_NAME(dcopy)(&pDetRE, REAL(sigmaSqPStarting_r), &inc, sigmaSqP, &inc); 
+    // Latent detection random effects
+    double *alphaStar = (double *) R_alloc(nDetRE, sizeof(double)); 
+    F77_NAME(dcopy)(&nDetRE, REAL(alphaStarStarting_r), &inc, alphaStar, &inc); 
     // Latent Occurrence
     double *z = (double *) R_alloc(J, sizeof(double));   
     F77_NAME(dcopy)(&J, REAL(zStarting_r), &inc, z, &inc);
@@ -133,12 +167,31 @@ extern "C" {
     PROTECT(zSamples_r = allocMatrix(REALSXP, J, nPost)); nProtect++; 
     SEXP psiSamples_r; 
     PROTECT(psiSamples_r = allocMatrix(REALSXP, J, nPost)); nProtect++; 
+    // Detection random effects
+    SEXP sigmaSqPSamples_r; 
+    SEXP alphaStarSamples_r; 
+    if (pDetRE > 0) {
+      PROTECT(sigmaSqPSamples_r = allocMatrix(REALSXP, pDetRE, nPost)); nProtect++;
+      PROTECT(alphaStarSamples_r = allocMatrix(REALSXP, nDetRE, nPost)); nProtect++;
+    }
+    // Occurrence random effects
+    SEXP sigmaSqPsiSamples_r; 
+    SEXP betaStarSamples_r; 
+    if (pOccRE > 0) {
+      PROTECT(sigmaSqPsiSamples_r = allocMatrix(REALSXP, pOccRE, nPost)); nProtect++;
+      PROTECT(betaStarSamples_r = allocMatrix(REALSXP, nOccRE, nPost)); nProtect++;
+    }
+    // Likelihood samples for WAIC. 
+    SEXP likeSamples_r;
+    PROTECT(likeSamples_r = allocMatrix(REALSXP, J, nPost)); nProtect++;
     
-    /**********************************************************************
-     * Other initial starting stuff
-     * *******************************************************************/
+    /********************************************************************
+      Some constants and temporary variables to be used later
+    ********************************************************************/
     int JpOcc = J * pOcc; 
     int nObspDet = nObs * pDet;
+    double tmp_0; 
+    double *tmp_one = (double *) R_alloc(inc, sizeof(double)); 
     double *tmp_ppDet = (double *) R_alloc(ppDet, sizeof(double));
     double *tmp_ppOcc = (double *) R_alloc(ppOcc, sizeof(double)); 
     double *tmp_pDet = (double *) R_alloc(pDet, sizeof(double));
@@ -150,19 +203,20 @@ extern "C" {
       tmp_J[j] = 0; 
     }
     double *tmp_nObs = (double *) R_alloc(nObs, sizeof(double)); 
-    zeros(tmp_nObs, nObs);
     double *tmp_JpOcc = (double *) R_alloc(JpOcc, sizeof(double));
-    zeros(tmp_JpOcc, JpOcc);
     double *tmp_nObspDet = (double *) R_alloc(nObspDet, sizeof(double));
-    zeros(tmp_nObspDet, nObspDet);
+    double *tmp_J1 = (double *) R_alloc(J, sizeof(double));
    
-    // For latent occupancy
+    // For latent occupancy and WAIC
     double psiNum; 
     double *detProb = (double *) R_alloc(nObs, sizeof(double)); zeros(detProb, nObs);
+    double *yWAIC = (double *) R_alloc(J, sizeof(double)); zeros(yWAIC, J);
     double *psi = (double *) R_alloc(J, sizeof(double)); 
     zeros(psi, J); 
     double *piProd = (double *) R_alloc(J, sizeof(double)); 
     ones(piProd, J); 
+    double *piProdWAIC = (double *) R_alloc(J, sizeof(double)); 
+    ones(piProdWAIC, J); 
     double *ySum = (double *) R_alloc(J, sizeof(double)); zeros(ySum, J);
 
     // For normal priors
@@ -172,15 +226,48 @@ extern "C" {
     F77_NAME(dpotri)(lower, &pOcc, SigmaBetaInv, &pOcc, &info); 
     if(info != 0){error("c++ error: dpotri SigmaBetaInv failed\n");}
     double *SigmaBetaInvMuBeta = (double *) R_alloc(pOcc, sizeof(double)); 
-    F77_NAME(dgemv)(ntran, &pOcc, &pOcc, &one, SigmaBetaInv, &pOcc, muBeta, &inc, &zero, SigmaBetaInvMuBeta, &inc); 	  
+    F77_NAME(dsymv)(lower, &pOcc, &one, SigmaBetaInv, &pOcc, muBeta, &inc, &zero, 
+        	    SigmaBetaInvMuBeta, &inc);
     // Detection regression coefficient priors. 
     F77_NAME(dpotrf)(lower, &pDet, SigmaAlphaInv, &pDet, &info); 
     if(info != 0){error("c++ error: dpotrf SigmaAlphaInv failed\n");}
     F77_NAME(dpotri)(lower, &pDet, SigmaAlphaInv, &pDet, &info); 
     if(info != 0){error("c++ error: dpotri SigmaAlphaInv failed\n");}
     double *SigmaAlphaInvMuAlpha = (double *) R_alloc(pDet, sizeof(double)); 
-    F77_NAME(dgemv)(ntran, &pDet, &pDet, &one, SigmaAlphaInv, &pDet, muAlpha, &inc, &zero, SigmaAlphaInvMuAlpha, &inc); 	  
+    F77_NAME(dsymv)(lower, &pDet, &one, SigmaAlphaInv, &pDet, muAlpha, &inc, &zero, 
+                   SigmaAlphaInvMuAlpha, &inc);
 
+    /**********************************************************************
+     * Prep for random effects
+     * *******************************************************************/
+    // Site-level sums of the occurrence random effects
+    double *betaStarSites = (double *) R_alloc(J, sizeof(double)); 
+    zeros(betaStarSites, J); 
+    // Initial sums
+    for (j = 0; j < J; j++) {
+      for (l = 0; l < pOccRE; l++) {
+        betaStarSites[j] += betaStar[which(XRE[l * J + j], betaLevelIndx, nOccRE)];
+      }
+    }
+    // Observation-level sums of the detection random effects
+    double *alphaStarObs = (double *) R_alloc(nObs, sizeof(double)); 
+    zeros(alphaStarObs, nObs); 
+    // Get sums of the current REs for each site/visit combo
+    for (i = 0; i < nObs; i++) {
+      for (l = 0; l < pDetRE; l++) {
+        alphaStarObs[i] += alphaStar[which(XpRE[l * nObs + i], alphaLevelIndx, nDetRE)];
+      }
+    }
+    // Starting index for occurrence random effects
+    int *betaStarStart = (int *) R_alloc(pOccRE, sizeof(int)); 
+    for (l = 0; l < pOccRE; l++) {
+      betaStarStart[l] = which(l, betaStarIndx, nOccRE); 
+    }
+    // Starting index for detection random effects
+    int *alphaStarStart = (int *) R_alloc(pDetRE, sizeof(int)); 
+    for (l = 0; l < pDetRE; l++) {
+      alphaStarStart[l] = which(l, alphaStarIndx, nDetRE); 
+    }
 
     GetRNGstate(); 
 
@@ -189,7 +276,7 @@ extern "C" {
        *Update Occupancy Auxiliary Variables 
        *******************************************************************/
       for (j = 0; j < J; j++) {
-        omegaOcc[j] = rpg(1.0, F77_NAME(ddot)(&pOcc, &X[j], &J, beta, &inc));
+        omegaOcc[j] = rpg(1.0, F77_NAME(ddot)(&pOcc, &X[j], &J, beta, &inc) + betaStarSites[j]);
       } // j
 
       /********************************************************************
@@ -197,11 +284,17 @@ extern "C" {
        *******************************************************************/
       if (nObs == J) {
         for (i = 0; i < nObs; i++) {
-          omegaDet[i] = rpg(K[i], F77_NAME(ddot)(&pDet, &Xp[i], &nObs, alpha, &inc));
+          if (z[zLongIndx[i]] == 1.0) {
+            omegaDet[i] = rpg(K[i], F77_NAME(ddot)(&pDet, &Xp[i], &nObs, alpha, &inc) + 
+        		      alphaStarObs[i]);
+          }
         } // i
       } else {
         for (i = 0; i < nObs; i++) {
-          omegaDet[i] = rpg(1.0, F77_NAME(ddot)(&pDet, &Xp[i], &nObs, alpha, &inc));
+          if (z[zLongIndx[i]] == 1.0) {
+            omegaDet[i] = rpg(1.0, F77_NAME(ddot)(&pDet, &Xp[i], &nObs, alpha, &inc) + 
+        		      alphaStarObs[i]);
+          }
         } // i
       }
 
@@ -209,13 +302,14 @@ extern "C" {
        *Update Occupancy Regression Coefficients
        *******************************************************************/
       for (j = 0; j < J; j++) {
-	kappaOcc[j] = z[j] - 1.0 / 2.0; 
+        kappaOcc[j] = z[j] - 1.0 / 2.0; 
+        tmp_J1[j] = kappaOcc[j] - omegaOcc[j] * betaStarSites[j]; 
       } // j
 
       /********************************
        * Compute b.beta
        *******************************/
-      F77_NAME(dgemv)(ytran, &J, &pOcc, &one, X, &J, kappaOcc, &inc, &zero, tmp_pOcc, &inc); 	 
+      F77_NAME(dgemv)(ytran, &J, &pOcc, &one, X, &J, tmp_J1, &inc, &zero, tmp_pOcc, &inc); 	 
       for (j = 0; j < pOcc; j++) {
         tmp_pOcc[j] += SigmaBetaInvMuBeta[j]; 
       } // j 
@@ -249,22 +343,24 @@ extern "C" {
       /********************************************************************
        *Update Detection Regression Coefficients
        *******************************************************************/
-      // /********************************
-      //  * Compute b.alpha
-      //  *******************************/
-      // First multiply kappDet * the current occupied values, such that values go 
-      // to 0 if z == 0 and values go to kappaDet if z == 1
+      /********************************
+       * Compute b.alpha
+       *******************************/
       if (nObs == J) {
         for (i = 0; i < nObs; i++) {
           kappaDet[i] = (y[i] - K[i]/2.0) * z[zLongIndx[i]];
+          tmp_nObs[i] = kappaDet[i] - omegaDet[i] * alphaStarObs[i]; 
+          tmp_nObs[i] *= z[zLongIndx[i]]; 
         } // i
-      } else { 
+      } else {
         for (i = 0; i < nObs; i++) {
           kappaDet[i] = (y[i] - 1.0/2.0) * z[zLongIndx[i]];
+          tmp_nObs[i] = kappaDet[i] - omegaDet[i] * alphaStarObs[i]; 
+          tmp_nObs[i] *= z[zLongIndx[i]]; 
         } // i
       }
       
-      F77_NAME(dgemv)(ytran, &nObs, &pDet, &one, Xp, &nObs, kappaDet, &inc, &zero, tmp_pDet, &inc); 	  
+      F77_NAME(dgemv)(ytran, &nObs, &pDet, &one, Xp, &nObs, tmp_nObs, &inc, &zero, tmp_pDet, &inc); 	  
       for (j = 0; j < pDet; j++) {
         tmp_pDet[j] += SigmaAlphaInvMuAlpha[j]; 
       } // j
@@ -294,37 +390,134 @@ extern "C" {
       if(info != 0){error("c++ error: dpotrf here failed\n");}
       mvrnorm(alpha, tmp_pDet2, tmp_ppDet, pDet);
 
+      /********************************************************************
+       *Update Occupancy random effects variance
+       *******************************************************************/
+      for (l = 0; l < pOccRE; l++) {
+        tmp_0 = F77_NAME(ddot)(&nOccRELong[l], &betaStar[betaStarStart[l]], &inc, &betaStar[betaStarStart[l]], &inc); 
+        tmp_0 *= 0.5; 
+        sigmaSqPsi[l] = rigamma(sigmaSqPsiA[l] + nOccRELong[l] / 2.0, sigmaSqPsiB[l] + tmp_0); 
+      }
+
+      /********************************************************************
+       *Update Detection random effects variance
+       *******************************************************************/
+      for (l = 0; l < pDetRE; l++) {
+        tmp_0 = F77_NAME(ddot)(&nDetRELong[l], &alphaStar[alphaStarStart[l]], &inc, &alphaStar[alphaStarStart[l]], &inc); 
+        tmp_0 *= 0.5; 
+        sigmaSqP[l] = rigamma(sigmaSqPA[l] + nDetRELong[l] / 2.0, sigmaSqPB[l] + tmp_0); 
+      }
+
+      /********************************************************************
+       *Update Occupancy random effects
+       *******************************************************************/
+      if (pOccRE > 0) {
+        // Update each individual random effect one by one. 
+        for (l = 0; l < nOccRE; l++) {
+          /********************************
+           * Compute b.beta.star
+           *******************************/
+          zeros(tmp_one, inc);
+          tmp_0 = 0.0;	      
+          // Only allow information to come from when XRE == betaLevelIndx[l]. 
+          // aka information only comes from the sites with any given level 
+          // of a random effect. 
+          for (j = 0; j < J; j++) {
+            if (XRE[betaStarIndx[l] * J + j] == betaLevelIndx[l]) {
+              tmp_one[0] += kappaOcc[j] - (F77_NAME(ddot)(&pOcc, &X[j], &J, beta, &inc) + 
+        		    betaStarSites[j] - betaStar[l]) * omegaOcc[j];
+              tmp_0 += omegaOcc[j];
+            }
+          }
+          /********************************
+           * Compute A.beta.star
+           *******************************/
+          tmp_0 += 1.0 / sigmaSqPsi[betaStarIndx[l]]; 
+          tmp_0 = 1.0 / tmp_0; 
+          betaStar[l] = rnorm(tmp_0 * tmp_one[0], sqrt(tmp_0)); 
+        }
+      
+        // Update the RE sums for the current species
+        zeros(betaStarSites, J);
+        for (j = 0; j < J; j++) {
+          for (l = 0; l < pOccRE; l++) {
+            betaStarSites[j] += betaStar[which(XRE[l * J + j], betaLevelIndx, nOccRE)];
+          }
+        }
+      }
+
+      /********************************************************************
+       *Update Detection random effects
+       *******************************************************************/
+      if (pDetRE > 0) {
+        // Update each individual random effect one by one. 
+        for (l = 0; l < nDetRE; l++) {
+          /********************************
+           * Compute b.alpha.star
+           *******************************/
+          // Only allow information to come from when z[r] == 1 and XpRE == alphaLevelIndx[l]
+          zeros(tmp_one, inc);
+          tmp_0 = 0.0;
+          for (r = 0; r < nObs; r++) {
+            if ((z[zLongIndx[r]] == 1.0) && (XpRE[alphaStarIndx[l] * nObs + r] == alphaLevelIndx[l])) {
+              tmp_one[0] += kappaDet[r] - (F77_NAME(ddot)(&pDet, &Xp[r], &nObs, alpha, &inc) + alphaStarObs[r] - alphaStar[l]) * omegaDet[r];
+      	      tmp_0 += omegaDet[r];
+            }
+          }
+          /********************************
+           * Compute A.alpha.star
+           *******************************/
+          tmp_0 += 1.0 / sigmaSqP[alphaStarIndx[l]]; 
+          tmp_0 = 1.0 / tmp_0; 
+          alphaStar[l] = rnorm(tmp_0 * tmp_one[0], sqrt(tmp_0)); 
+        }
+        zeros(alphaStarObs, nObs); 
+        // Update the RE sums for the current species
+        for (r = 0; r < nObs; r++) {
+          for (l = 0; l < pDetRE; l++) {
+            alphaStarObs[r] += alphaStar[which(XpRE[l * nObs + r], alphaLevelIndx, nDetRE)]; 
+          }
+        }
+      }
       /********************************************************************
        *Update Latent Occupancy
        *******************************************************************/
       // Compute detection probability 
       if (nObs == J) {
         for (i = 0; i < nObs; i++) {
-          detProb[i] = logitInv(F77_NAME(ddot)(&pDet, &Xp[i], &nObs, alpha, &inc), zero, one);
-          psi[zLongIndx[i]] = logitInv(F77_NAME(ddot)(&pOcc, &X[zLongIndx[i]], &J, beta, &inc), zero, one); 
+          detProb[i] = logitInv(F77_NAME(ddot)(&pDet, &Xp[i], &nObs, alpha, &inc) + alphaStarObs[i], zero, one);
+          psi[zLongIndx[i]] = logitInv(F77_NAME(ddot)(&pOcc, &X[zLongIndx[i]], &J, beta, &inc) + betaStarSites[zLongIndx[i]], zero, one); 
           piProd[zLongIndx[i]] = pow(1.0 - detProb[i], K[i]);
-	  ySum[zLongIndx[i]] = y[i]; 
+	  piProdWAIC[zLongIndx[i]] *= pow(detProb[i], y[i]);
+	  piProdWAIC[zLongIndx[i]] *= pow(1.0 - detProb[i], K[i] - y[i]);
+          ySum[zLongIndx[i]] = y[i]; 	
         } // i
       } else {
         for (i = 0; i < nObs; i++) {
-          detProb[i] = logitInv(F77_NAME(ddot)(&pDet, &Xp[i], &nObs, alpha, &inc), zero, one);
+          detProb[i] = logitInv(F77_NAME(ddot)(&pDet, &Xp[i], &nObs, alpha, &inc) + alphaStarObs[i], zero, one);
           if (tmp_J[zLongIndx[i]] == 0) {
-            psi[zLongIndx[i]] = logitInv(F77_NAME(ddot)(&pOcc, &X[zLongIndx[i]], &J, beta, &inc), zero, one); 
+            psi[zLongIndx[i]] = logitInv(F77_NAME(ddot)(&pOcc, &X[zLongIndx[i]], &J, beta, &inc) + betaStarSites[zLongIndx[i]], zero, one); 
           }
           piProd[zLongIndx[i]] *= (1.0 - detProb[i]);
+	  piProdWAIC[zLongIndx[i]] *= pow(detProb[i], y[i]);
+	  piProdWAIC[zLongIndx[i]] *= pow(1.0 - detProb[i], 1 - y[i]);
           ySum[zLongIndx[i]] += y[i]; 	
           tmp_J[zLongIndx[i]]++;
         } // i
       }
-      // Compute occupancy probability 
+
+      // Compute occupancy probability and the integrated likelihood for WAIC
       for (j = 0; j < J; j++) {
         psiNum = psi[j] * piProd[j]; 
         if (ySum[j] == zero) {
-          z[j] = rbinom(one, psiNum / (psiNum + (1.0 - psi[j])));           
+          z[j] = rbinom(one, psiNum / (psiNum + (1.0 - psi[j])));          
+	  yWAIC[j] = (1.0 - psi[j]) + psi[j] * piProdWAIC[j]; 
         } else {
           z[j] = one; 
+	  yWAIC[j] = psi[j] * piProdWAIC[j];
         }
         piProd[j] = one;
+	piProdWAIC[j] = one;
         ySum[j] = zero; 
         tmp_J[j] = 0; 
       } // j
@@ -334,14 +527,28 @@ extern "C" {
        *******************************************************************/
       if (s >= nBurn) {
         thinIndx++; 
-	if (thinIndx == nThin) {
+        if (thinIndx == nThin) {
           F77_NAME(dcopy)(&pOcc, beta, &inc, &REAL(betaSamples_r)[sPost*pOcc], &inc);
           F77_NAME(dcopy)(&pDet, alpha, &inc, &REAL(alphaSamples_r)[sPost*pDet], &inc);
           F77_NAME(dcopy)(&J, psi, &inc, &REAL(psiSamples_r)[sPost*J], &inc); 
           F77_NAME(dcopy)(&J, z, &inc, &REAL(zSamples_r)[sPost*J], &inc); 
+          if (pOccRE > 0) {
+            F77_NAME(dcopy)(&pOccRE, sigmaSqPsi, &inc, 
+        		    &REAL(sigmaSqPsiSamples_r)[sPost*pOccRE], &inc);
+            F77_NAME(dcopy)(&nOccRE, betaStar, &inc, 
+        		    &REAL(betaStarSamples_r)[sPost*nOccRE], &inc);
+          }
+          if (pDetRE > 0) {
+            F77_NAME(dcopy)(&pDetRE, sigmaSqP, &inc, 
+        		    &REAL(sigmaSqPSamples_r)[sPost*pDetRE], &inc);
+            F77_NAME(dcopy)(&nDetRE, alphaStar, &inc, 
+        		    &REAL(alphaStarSamples_r)[sPost*nDetRE], &inc);
+          }
+          F77_NAME(dcopy)(&J, yWAIC, &inc, 
+        		  &REAL(likeSamples_r)[sPost*J], &inc);
           sPost++; 
-	  thinIndx = 0; 
-	}
+          thinIndx = 0; 
+        }
       }
 
       /********************************************************************
@@ -368,7 +575,13 @@ extern "C" {
     PutRNGstate();
 
     SEXP result_r, resultName_r;
-    int nResultListObjs = 4;
+    int nResultListObjs = 5;
+    if (pDetRE > 0) {
+      nResultListObjs += 2; 
+    }
+    if (pOccRE > 0) {
+      nResultListObjs += 2;
+    }
 
     PROTECT(result_r = allocVector(VECSXP, nResultListObjs)); nProtect++;
     PROTECT(resultName_r = allocVector(VECSXP, nResultListObjs)); nProtect++;
@@ -378,11 +591,34 @@ extern "C" {
     SET_VECTOR_ELT(result_r, 1, alphaSamples_r);
     SET_VECTOR_ELT(result_r, 2, zSamples_r); 
     SET_VECTOR_ELT(result_r, 3, psiSamples_r);
+    SET_VECTOR_ELT(result_r, 4, likeSamples_r);
+    if (pDetRE > 0) {
+      SET_VECTOR_ELT(result_r, 5, sigmaSqPSamples_r);
+      SET_VECTOR_ELT(result_r, 6, alphaStarSamples_r);
+    }
+    if (pOccRE > 0) {
+      if (pDetRE > 0) {
+        tmp_0 = 7;
+      } else {
+        tmp_0 = 5;
+      }
+      SET_VECTOR_ELT(result_r, tmp_0, sigmaSqPsiSamples_r);
+      SET_VECTOR_ELT(result_r, tmp_0 + 1, betaStarSamples_r);
+    }
 
     SET_VECTOR_ELT(resultName_r, 0, mkChar("beta.samples")); 
     SET_VECTOR_ELT(resultName_r, 1, mkChar("alpha.samples")); 
     SET_VECTOR_ELT(resultName_r, 2, mkChar("z.samples")); 
     SET_VECTOR_ELT(resultName_r, 3, mkChar("psi.samples"));
+    SET_VECTOR_ELT(resultName_r, 4, mkChar("like.samples"));
+    if (pDetRE > 0) {
+      SET_VECTOR_ELT(resultName_r, 5, mkChar("sigma.sq.p.samples")); 
+      SET_VECTOR_ELT(resultName_r, 6, mkChar("alpha.star.samples")); 
+    }
+    if (pOccRE > 0) {
+      SET_VECTOR_ELT(resultName_r, tmp_0, mkChar("sigma.sq.psi.samples")); 
+      SET_VECTOR_ELT(resultName_r, tmp_0 + 1, mkChar("beta.star.samples")); 
+    }
    
     namesgets(result_r, resultName_r);
     
