@@ -17,6 +17,7 @@ extern "C" {
   SEXP spMsPGOccPredict(SEXP J_r, SEXP N_r, SEXP pOcc_r, SEXP X0_r, SEXP q_r, 
 		      SEXP obsD_r, SEXP obsPredD_r, SEXP betaSamples_r, 
 		      SEXP thetaSamples_r, SEXP wSamples_r, 
+		      SEXP betaStarSiteSamples_r,
 		      SEXP nSamples_r, SEXP covModel_r, SEXP nThreads_r, 
 		      SEXP verbose_r, SEXP nReport_r){
     
@@ -51,6 +52,7 @@ extern "C" {
     double *betaSamples = REAL(betaSamples_r);
     double *thetaSamples = REAL(thetaSamples_r);
     double *wSamples = REAL(wSamples_r);
+    double *betaStarSite = REAL(betaStarSiteSamples_r);
     
     int nSamples = INTEGER(nSamples_r)[0];
     int covModel = INTEGER(covModel_r)[0];
@@ -137,11 +139,11 @@ extern "C" {
         phi = thetaSamples[s * nThetaN + phiIndx * N + i]; 
         if (corName == "matern") {
           nu = thetaSamples[s * nThetaN + nuIndx * N + i]; 
+          theta[nuIndx] = nu; 
         }
         sigmaSq = thetaSamples[s * nThetaN + sigmaSqIndx * N + i]; 
         theta[sigmaSqIndx] = sigmaSq; 
         theta[phiIndx] = phi; 
-        theta[nuIndx] = nu; 
 
         // Get covariance matrices
         spCov(obsD, JJ, theta, corName, S_obs); 
@@ -166,7 +168,7 @@ extern "C" {
           tmp_one2[0] = sigmaSq - tmp_one2[0];
 
           w0[s * qN + j * N + i] = rnorm(tmp_one[0], sqrt(tmp_one2[0])); 
-          psi0[s * qN + j * N + i] = logitInv(tmp_q[j] + w0[s * qN + j * N + i], zero, one); 
+          psi0[s * qN + j * N + i] = logitInv(tmp_q[j] + w0[s * qN + j * N + i] + betaStarSite[s * qN + j * N + i], zero, one); 
           z0[s * qN + j * N + i] = rbinom(one, psi0[s * qN + j * N + i]);
         }
 
